local AddonName = ...
local Addon = _G[AddonName]

local kinds = {}
local kinds_mt = setmetatable({}, {
    __index = function(self, k)
        local kind = kinds[k]

        if kind then
            local mt = { __index = kind }

            self[k] = mt

            return mt
        end
    end
})

local slots = {}
local function slotIterator(filter, i)
    local slot
    repeat
        i, slot = next(slots, i)
        if i ~= nil and filter(slot) then
            return i, slot
        end
    until i == nil
end

local ActionSlots = Addon:NewModule("ActionSlots", "AceConsole-3.0")

function ActionSlots:OnEnable()
    Addon.RegisterCallback(self, "OnActionSlotPickupAction")
end

function ActionSlots:OnActionSlotPickupAction(msg, slot, kind, action)
    ClearCursor()
end

function ActionSlots:Create(id)
    id = tonumber(id)

    if not id then
        error("Usage ActionSlots:Create(id)", 2)
    end

    local name = ("%sActionButton%d"):format(AddonName, id)
    if _G[name] ~= nil then
        error(("ActionSlot %d already exists"):format(id), 2)
    end

    local frame = CreateFrame("CheckButton", name, nil, "SecureActionButtonTemplate, ActionButtonTemplate")

    local slot = setmetatable(frame, kinds_mt.basic)

    slot.id = id
    slot:OnLoad()

    slots[id] = slot
    return slot
end

function ActionSlots:SetKind(slot, kind)
    if not kinds[kind] then
        error(("ActionSlotKind %q not yet registered"):format(kind), 3)
    end

    if slot.kind ~= kind then
        setmetatable(slot, kinds_mt[kind])
        return true
    end
end

function ActionSlots:Get(id)
    return slots[id]
end

function ActionSlots:GetOrCreate(id)
    return self:Get(id) or self:Create(id)
end

function ActionSlots:GetAll()
    return next, slots, nil
end

function ActionSlots:GetAll(kind)
    return self:GetFiltered(function(s) return s.kind == kind end)
end

function ActionSlots:GetFiltered(func, ...)
    if type(func) == "function" then
        return slotIterator, func, nil
    end

    if type(func) == "string" then
        return slotIterator, function(s) return s[func](s) end, nil
    end

    error('Usage: ActionSlots:GetFiltered(function or "method")', 2)
end

function ActionSlots:CreateKind(kind, prototype)
    if type(kind) ~= "string" then
        error('Usage: ActionSlots:CreateKind("kind", ["prototype"])', 2)
    end

    if kinds[kind] then
        error(("ActionSlot kind %q is already registered"):format(kind), 2)
    end

    if prototype and type(prototype) ~= "string" then
        error('Usage: ActionSlots:CreateKind("kind", ["prototype"])', 2)
    elseif not prototype then
        prototype = "basic"
    end

    -- hack to handle basic kind creation
    local result
    if kind == "basic" then
        local f = CreateFrame("CheckButton", nil, nil, "SecureHandlerBaseTemplate")

        f:Hide()

        result = setmetatable({ kind = kind }, {__index = f })
    else
        local metatable = kinds_mt[prototype]

        if not metatable then
            error(("ActionSlotKind %q not yet registered"):format(prototype), 2)
        end

        result = setmetatable({ kind = kind }, metatable)
    end

    kinds[kind] = result
    return result
end

function ActionSlots:GetKind(kind)
    if type(kind) ~= "string" then
        error('Usage: ActionSlots:GetKind("kind")', 2)
    end

    return kinds[kind]
end

Addon.ActionSlots = ActionSlots