local Addon = _G[...]
local GetCursorInfo = _G.GetCursorInfo
local GetLastNonNilIndex = Addon.Util.GetLastNonNilIndex
local ToFunctionArgs = Addon.Util.ToFunctionArgs

local function secureHandler_Create(owner)
    local frame = CreateFrame("Frame", nil, nil, "SecureHandlerStateTemplate")

    frame:SetScript(
        "OnEvent",
        function(_, event, ...)
            local func = owner[event]
            if type(func) == "function" then
                func(owner, event, ...)
            end
        end
    )

    frame:Hide()

    return frame
end

local function customCursor_Create(owner)
    local parent = _G.UIParent

    local cursor = CreateFrame("Frame", nil, parent)
    cursor:SetFrameStrata("DIALOG")
    cursor:SetToplevel(true)
    cursor:Hide()

    local icon = cursor:CreateTexture(nil, "OVERLAY")
    icon:SetSize(26, 26)
    cursor.icon = icon

    local modal = CreateFrame("Button", nil, parent)
    modal:SetFrameStrata("LOW")
    modal:SetAllPoints(parent)
    modal:RegisterForClicks("AnyDown")
    modal:SetScript("OnClick", function() ClearCursor() end)
    modal:Hide()

    cursor:SetScript("OnShow", function()
        modal:Show()
    end)

    cursor:SetScript("OnHide", function()
        modal:Hide()
    end)

    cursor:SetScript("OnUpdate", function()
        local x, y = GetCursorPosition()
        local scale = parent:GetScale()
        icon:SetPoint("CENTER", parent, "BOTTOMLEFT", x / scale, y / scale)
    end)

    return cursor
end

-- hook all of the known methods for mutating cursor info
local function hookCursorFunctions(owner)
    for i, v in pairs(_G) do
        -- handle global pickup functions (ex, PickupAction)
        if type(v) == "function" and i:match("^Pickup") then
            -- handle C_Library pickup functions
            hooksecurefunc(
                i,
                function(...)
                    owner:OnPickup(i, ...)
                end
            )
        elseif type(v) == "table" and i:match("C_%w+") then
            for j, vv in pairs(v) do
                if type(vv) == "function" and j:match("^Pickup") then
                    hooksecurefunc(
                        v,
                        j,
                        function(...)
                            owner:OnPickup(strjoin(".", i, j), ...)
                        end
                    )
                end
            end
        end
    end

    hooksecurefunc(
        "ClearCursor",
        function()
            owner:ClearInfo()
        end
    )
end

local SecureEnv = {}

function SecureEnv:OnLoad()
    self.actionSlots = {}
    self.cursorInfo = {}

    self.frame = secureHandler_Create(self)

    self.frame:Execute([[
        CursorInfo = table.new()
        Frames = table.new()
    ]])

    self.frame:SetAttribute(
        "SetCursorInfo",
        [[
            wipe(CursorInfo)

            for i = 1, select("#", ...) do
                CursorInfo[i] = select(i, ...)
            end
        ]]
    )

    self.frame:SetAttribute("GetCursorInfo", [[ return unpack(CursorInfo) ]])
    self.frame:SetAttribute("HasCursorInfo", [[ return #CursorInfo > 0 ]])

    hookCursorFunctions(self)

    Addon.RegisterCallback(self, "OnBarCreated")
    Addon.RegisterCallback(self, "OnActionSlotActivated")

    self.frame:RegisterEvent("CURSOR_UPDATE")
    self.frame:RegisterEvent("PLAYER_LOGIN")
    self.frame:RegisterEvent("PLAYER_REGEN_ENABLED")
    self.frame:RegisterEvent("PLAYER_REGEN_DISABLED")

    self.frame:RegisterEvent("PET_BAR_UPDATE")
    self.frame:RegisterEvent("PET_UI_UPDATE")
    self.frame:RegisterEvent("UPDATE_VEHICLE_ACTIONBAR")
    self.frame:RegisterUnitEvent("UNIT_PET", "player")
    self.frame:RegisterEvent("ACTIONBAR_SLOT_CHANGED")
    self.frame:RegisterEvent("SPELLS_CHANGED")

    self.cursor = customCursor_Create(self)

    hooksecurefunc("ClearCursor", function() self:ClearCustom() end)
end

function SecureEnv:CURSOR_UPDATE()
    if not GetCursorInfo() then
        self:SetCursorInfo()
    end
end

function SecureEnv:PLAYER_LOGIN()
    self:UpdateActionSlots()
    self:UpdatePetActionSlots()
end

function SecureEnv:ACTIONBAR_SLOT_CHANGED(msg, slot)
    self:UpdateActionSlot(slot)
end

function SecureEnv:PLAYER_REGEN_ENABLED()
    self.inCombat = nil
end

function SecureEnv:PLAYER_REGEN_DISABLED()
    self.inCombat = true

    if self.cursorNeedsUpdate then
        self:UpdateSecureCursorInfo()
    end
end

function SecureEnv:PET_BAR_UPDATE()
    self:UpdatePetActionSlots()
end

function SecureEnv:PET_UI_UPDATE()
    self:UpdatePetActionSlots()
end

function SecureEnv:UNIT_PET()
    self:UpdatePetActionSlots()
end

function SecureEnv:UPDATE_VEHICLE_ACTIONBAR()
    self:UpdatePetActionSlots()
end

function SecureEnv:SPELLS_CHANGED()
    self:UpdatePetActionSlots()
end

function SecureEnv:OnActionSlotActivated(msg, slot)
    self:AddActionSlot(slot)
end

function SecureEnv:OnBarCreated(msg, bar)
    self.frame:SetFrameRef("addFrame", bar.frame)

    self.frame:Execute[[
        local f = self:GetFrameRef("addFrame")

        Frames[f] = true
    ]]
end

function SecureEnv:OnPickup(name, ...)
    if name == "PickupAction" then
        self:OnPickupAction(...)
    elseif name == "PickupPetAction" then
        self:OnPickupPetAction(...)
    elseif name == "PickupSpellBookItem" then
        self:OnPickupSpellBookItem(...)
    else
        self:SetCursorInfo(GetCursorInfo())
    end
end

function SecureEnv:OnPickupAction(slot)
    local id = self:GetActionSlotPrevPetActionID(slot)
    if id then
        self:PickupPetActionID(id)
    else
        self:SetCursorInfo(GetCursorInfo())
    end
end

function SecureEnv:OnPickupPetAction(slot)
    local id = self:GetPrevPetActionID(slot)
    if id then
        self:PickupPetActionID(id)
    else
        self:SetCursorInfo(GetCursorInfo())
    end
end

function SecureEnv:OnPickupSpellBookItem(index, bookType)
    if bookType == BOOKTYPE_PET then
        local _, petActionID = GetSpellBookItemInfo(index, bookType)
        self:PickupPetActionID(petActionID)
    else
        self:SetCursorInfo(GetCursorInfo())
    end
end

function SecureEnv:PickupAction(kind, ...)
    ClearCursor()

	if kind == "clear" or kind == "empty" then
        return
	end

    if kind == 'action' then
        local slot = ...
        return _G.PickupAction(slot)
    end

    if kind == 'actionpage' then
        local pageID = ...
        return Addon.ActionPageSlot.Pickup(pageID)
    end

    if kind == 'battlepet' then
        local battlePetID = ...
        return Addon.BattlePetSlot.Pickup(battlePetID)
    end

    if kind == 'equipmentset' then
        local setName = ...
        return Addon.EquipmentSetSlot.Pickup(setName)
    end

    if kind == 'exitvehicle' then
        return Addon.ExitVehicleSlot.Pickup()
    end

    if kind == 'item' then
        local itemString = ...
        return _G.PickupItem(itemString)
    end

    if kind == 'flyout' then
        local flyoutID = ...
        return Addon.FlyoutSlot.Pickup(flyoutID)
    end

    if kind == 'macro' then
        local macroID = ...
        return _G.PickupMacro(macroID)
    end

    if kind == 'mount' then
        local mountID = ...
        return Addon.MountSlot.Pickup(mountID)
    end

    if kind == 'movement' then
        local movementAction = ...
        return Addon.MovementSlot.Pickup(movementAction)
    end

    if kind == 'overrideaction' then
        local actionID = ...
        return Addon.OverrideActionSlot.Pickup(actionID)
    end

    if kind == 'petability' or kind == 'petcommand' then
        local petActionID = strsplit("|", ...)
        return Addon.PetActions.Pickup(tonumber(petActionID))
    end

    if kind == 'spell' then
        local spellID = ...
        return _G.PickupSpell(spellID)
    end
end

function SecureEnv:PickupPetActionID(id)
    local name, kind, _, spellID, command = Addon.PetActions:GetPetActionIDInfo(id)

    if spellID then
        self:SetCursorInfo("petability", id, name, kind)
    else
        self:SetCursorInfo("petcommand", id, command)
    end
end

function SecureEnv:PickupCustom(kind, action, icon)
    self:SetCursorInfo(kind, action)

    self.cursor.icon:SetTexture(icon)
    self.cursor:Show()

    PlaySound(832) -- IG_ABILITY_ICON_PICKUP
end

function SecureEnv:ClearCustom()
    if self.cursor:IsShown() then
        self.cursor:Hide()
        self.cursor.icon:SetTexture(nil)
        PlaySound(SOUNDKIT.IG_ABILITY_ICON_DROP)

        self:SetCursorInfo()
    end
end

function SecureEnv:SetCursorInfo(...)
    local hadInfo = next(self.cursorInfo) and true or false

    wipe(self.cursorInfo)

    local count = GetLastNonNilIndex(...)
    if count > 0 then
        for i = 1, count do
            self.cursorInfo[i] = select(i, ...)
        end
    end

    local hasInfo = next(self.cursorInfo) and true or false

    if hasInfo ~= hadInfo then
        if hasInfo then
            Addon:Notify("OnCursorPickupAction")
        else
            Addon:Notify("OnCursorDropAction")
        end
    end

    self:UpdateSecureCursorInfo()
end

function SecureEnv:ClearInfo()
    wipe(self.cursorInfo)
    self:UpdateSecureCursorInfo()
end

function SecureEnv:GetCursorInfo()
    return unpack(self.cursorInfo)
end

function SecureEnv:HasCursorInfo()
    return #self.cursorInfo > 0
end

function SecureEnv:UpdateSecureCursorInfo()
    if self.inCombat then
        self.cursorNeedsUpdate = true
        return
    end

    if self:HasCursorInfo() then
        local args = ToFunctionArgs(self:GetCursorInfo())
        local cmd = ([[ self:RunAttribute("SetCursorInfo", %s) ]]):format(args)

        self.frame:Execute(cmd)
    else
        self.frame:Execute([[ self:RunAttribute("SetCursorInfo") ]])
    end

    self.cursorNeedsUpdate = nil
end

do
    local NUM_PET_ACTION_SLOTS = _G.NUM_PET_ACTION_SLOTS
    local PetActions = Addon.PetActions
    local PetHasActionBar = _G.PetHasActionBar

    local info = {}
    local prevInfo = {}

    function SecureEnv:GetPrevPetActionID(slot)
        return prevInfo[slot]
    end

    function SecureEnv:UpdatePetActionSlots()
        if not PetHasActionBar() then
            wipe(info)
            wipe(prevInfo)
        elseif not next(prevInfo) then
            for i = 1, NUM_PET_ACTION_SLOTS do
                local id = PetActions:GetPetActionIDFromPetActionSlot(i)
                prevInfo[i] = id
                info[i] = id
            end
        else
            for i = 1, NUM_PET_ACTION_SLOTS do
                prevInfo[i] = info[i]
                info[i] = PetActions:GetPetActionIDFromPetActionSlot(i)
            end
        end
    end
end

do
    local PetActions = Addon.PetActions
    local NUM_ACTION_SLOTS = 120
    local info, prevInfo

    function SecureEnv:GetActionSlotPrevPetActionID(slot)
        if prevInfo then
            return prevInfo[slot]
        end
    end

    function SecureEnv:UpdateActionSlots()
        if not info then
            info = {}
            prevInfo = {}

            for i = 1, NUM_ACTION_SLOTS do
                local id = PetActions:GetPetActionIDFromActionSlot(i)
                prevInfo[i] = id
                info[i] = id
            end
        else
            for i = 1, NUM_ACTION_SLOTS do
                self:UpdateActionSlot(i)
            end
        end
    end

    function SecureEnv:UpdateActionSlot(slot)
        if info then
            prevInfo[slot] = info[slot]
            info[slot] = PetActions:GetPetActionIDFromActionSlot(slot)
        end
    end
end

function SecureEnv:AddActionSlot(slot)
    if self.actionSlots[slot] then
        return
    end

    self.actionSlots[slot] = true

    self.frame:WrapScript(
        slot,
        "OnClick",
        [[
            local env = self:GetFrameRef("SecureEnv")
            local handle = (self:GetAttribute("lockactionbar") and IsModifiedClick("PICKUPACTION")) or env:RunAttribute("HasCursorInfo")

            if handle and self:IsUnderMouse() then
                self:CallMethod("OnDrop", env:RunAttribute("GetCursorInfo"))
                return false
            end
        ]]
    )

    self.frame:WrapScript(
        slot,
        "OnDragStart",
        [[
            local handle = IsModifiedClick("PICKUPACTION") or (not self:GetAttribute("lockactionbar"))

            if handle then
                local env = self:GetFrameRef("SecureEnv")
                self:CallMethod("OnDrop", env:RunAttribute("GetCursorInfo"))
                return false
            end
        ]]
    )

    self.frame:WrapScript(
        slot,
        "OnReceiveDrag",
        [[
            local env = self:GetFrameRef("SecureEnv")

            if env:RunAttribute("HasCursorInfo") then
                self:CallMethod("OnDrop", env:RunAttribute("GetCursorInfo"))
                return false
            end
        ]]
    )

    slot:SetFrameRef("SecureEnv", self.frame)
end

SecureEnv:OnLoad()
Addon.SecureEnv = SecureEnv
