-- bar.lua
-- a generic container object
--
-- Callbacks:
-- OnBarCreated(bar)
-- OnBarDeactivated(bar)
-- OnBarReactivated(bar)
-- OnBarPropertyChanged(bar, settingName, ...)

local AddonName = ...
local Addon = _G[AddonName]
local Bar = {}

local active = {}
local inactive = {}

local function header_Create(owner)
	local name = ('%sBar%s'):format(AddonName, owner.id)

	local header = CreateFrame('Frame', name, _G.UIParent, 'SecureHandlerStateTemplate')

	header:SetAttribute('id', owner.id)

	return header
end

function Bar:GetOrCreate(id, name, description)
	return self:Get(id) or self:Create(id, name, description)
end

function Bar:Create(id, name, description)
	id = tonumber(id) or id

	if active[id] then
		error(("Bar %q is already active"):format(id), 2)
	end

	local bar = self:Activate(id) or self:Construct(id)
	bar.name = name
	bar.description = description
	active[id] = bar

	bar:OnCreate()
	Addon.Callbacks:Fire("OnBarCreated", bar)

	return bar
end

function Bar:Activate(id)
	local bar = inactive[id]

	if bar then
		inactive[id] = nil

		bar:OnActivate()
		Addon.Callbacks:Fire("OnBarActivated", bar)
		return bar
	end
end

--destructor
function Bar:Deactivate()
	if not active[self.id] then return end

	active[self.id] = nil
	inactive[self.id] = self

	self.frame:ClearAllPoints()
	self.frame:SetUserPlaced(nil)
	self.frame:Hide()

	UnregisterStateDriver(self.frame, 'display', 'show')

	self:OnDeactivate()
	Addon.Callbacks:Fire("OnBarDeactivated", self)
end

function Bar:Construct(id)
	local bar = Mixin({id = id}, self)

	local frame = header_Create(bar)
	frame:SetClampedToScreen(true)
	frame:SetMovable(true)
	bar.frame = frame

	bar:OnConstruct()
	Addon.Callbacks:Fire("OnBarConstructed", bar)

	return bar
end

function Bar:Initialize(settings)
	if type(settings) == "function" then
		self.sets = settings(self)
	elseif type(settings) == "table" then
		self.sets = settings
	else
		self.sets = {}
	end

	self:OnInitialize()
	Addon.Callbacks:Fire("OnBarInitialized", self)
end

-- called when a new bar is created
function Bar:OnCreate()
end

function Bar:OnActivate()
end

function Bar:OnDeactivate()
end

function Bar:OnConstruct()
end

function Bar:OnInitialize()
	self:UpdateShowStates()
	self:UpdateShown()
	self:UpdateAlpha()
	self:UpdatePosition()
end

--[[ Scaling ]]--

local function floatsEqual(x, y, precision)
	local mult = 10 ^ (precision or 2)

	return floor((x * mult) - (y * mult)) == 0
end

function Bar:SetScale(scale)
	local oldScale = self:GetScale()
	local newScale = tonumber(scale) or 1

	if not floatsEqual(oldScale, newScale) then
		self.sets.scale = scale

		local frame = self.frame
		local parent = frame:GetParent()
		local ratio = oldScale / newScale
		local x = (frame:GetLeft() or 0) * ratio
		local y = (frame:GetTop() or 0) * ratio

		frame:ClearAllPoints()
		frame:SetScale(newScale)
		frame:SetPoint('TOPLEFT', parent, 'BOTTOMLEFT', x, y)

		self:SaveRelativePosition()
		Addon.Callbacks:Fire("OnBarPropertyChanged", self, "scale", self:GetScale())
	end
end

function Bar:GetScale()
	return self.sets.scale or 1
end

function Bar:SetAlpha(alpha)
	local oldAlpha = self:GetAlpha()
	local newAlpha = tonumber(alpha) or 1

	if not floatsEqual(oldAlpha, newAlpha) then
		self.sets.alpha = newAlpha
		self:UpdateAlpha()

		Addon.Callbacks:Fire("OnBarPropertyChanged", self, "alpha", alpha)
	end
end

function Bar:GetAlpha()
	return self.sets.alpha or 1
end

function Bar:UpdateAlpha()
	local newAlpha = self:CalculateAlpha()
	local oldAlpha = self.frame:GetAlpha()

	if not floatsEqual(oldAlpha, newAlpha) then
		self.frame:SetAlpha(newAlpha)
	end
end

--faded opacity (mouse not over the frame)
function Bar:SetFadeMultiplier(multiplier)
	local oldMult = self:GetFadeMultiplier()
	local newMult = tonumber(multiplier) or 1

	if oldMult ~= newMult then
		self.sets.fadeMultiplier = newMult
		self:UpdateAlpha()

		Addon.Callbacks:Fire("OnBarPropertyChanged", self, "fadeMultiplier", self:GetFadeMultiplier())
	end
end

function Bar:GetFadeMultiplier()
	return self.sets.fadeMultiplier or 1
end

--returns the opacity value we expect the frame to be at in its current state
function Bar:CalculateAlpha()
	--if there's a statealpha value for the frame, then use it
	local alpha = self.frame:GetAttribute("state-alpha")
	if alpha then
		return alpha / 100
	end

	--if the frame is moused over, then return the frame's normal opacity
	if self.focused then
		return self:GetAlpha()
	end

	return self:GetAlpha() * self:GetFadeMultiplier()
end

-- visibility
function Bar:SetShown(show)
	local wasShown = self:IsShown()
	local isShown = show and true or false

	if isShown ~= wasShown then
		self.sets.shown = isShown

		self:UpdateShown()
		Addon.Callbacks:Fire("OnBarPropertyChanged", self, "show", isShown)
	end
end

function Bar:ToggleShown()
	self:SetShown(not self:IsShown())
end

function Bar:IsShown()
	return self.sets.shown and true or false
end

function Bar:UpdateShown()
	if self:IsShown() then
		self.frame:Show()
	else
		self.frame:Hide()
	end
end

-- perspectives
-- function Bar:SetShowInOverrideUI(show)
-- 	local wasShown = self:ShowingInOverrideUI()
-- 	local isShown = show and true or false

-- 	if isShown ~= wasShown then
-- 		self.sets.showInOverrideUI = isShown
-- 		self.frame:SetAttribute('state-showinoverrideui', isShown)
-- 		Addon.Callbacks:Fire("OnBarPropertyChanged", self, "showInOverrideUI", isShown)
-- 	end
-- end

-- function Bar:IsShownInOverrideUI()
-- 	return self.sets.showInOverrideUI and true or false
-- end

-- function Bar:SetShowInPetBattleUI(show)
-- 	local wasShown = self:ShowingInOverrideUI()
-- 	local isShown = show and true or false

-- 	if isShown ~= wasShown then
-- 		self.sets.showInPetBattleUI = isShown
-- 		self.frame:SetAttribute('state-showinpetbattleui', isShown)
-- 		Addon.Callbacks:Fire("OnBarPropertyChanged", self, "showInOverrideUI", isShown)
-- 	end
-- end

-- function Bar:IsShownInPetBattleUI()
-- 	return self.sets.showInPetBattleUI and true or false
-- end

-- showstates
function Bar:SetShowStates(states)
	local oldStates = self:GetShowStates()
	local newStates = tostring(states)

	if oldStates ~= newStates then
		self.sets.showStates = newStates
		self:UpateShowStates()
		Addon.Callbacks:Fire("OnBarPropertyChanged", self, "showStates", newStates)
	end
end

function Bar:GetShowStates()
	return self.sets.showStates
end

function Bar:UpdateShowStates()
	local showStates = self:GetShowStates()

	if showStates and showStates ~= '' then
		RegisterStateDriver(self.frame, 'display', showStates)
	else
		UnregisterStateDriver(self.frame, 'display')

		if self.frame:GetAttribute('state-display') then
			self.frame:SetAttribute('state-display', nil)
		end
	end
end

-- Movement
function Bar:StartMoving()
	self.frame:StartMoving()
end

function Bar:StopMovingOrSizing()
	self.frame:StopMovingOrSizing()
	self:SnapToGrid()
	self:SaveRelativePosition()
end

-- Positioning
function Bar:SetSavedPosition(point, x, y)
	local oldPoint, oldX, oldY = self:GetSavedPosition()
	local newPoint, newX, newY = point or "CENTER", Round(x), Round(y)

	if not (newPoint == oldPoint and floatsEqual(newX, oldX) and floatsEqual(newY, oldY)) then
		self.sets.point = newPoint
		self.sets.x = newX
		self.sets.y = newY

		Addon.Callbacks:Fire("OnBarPropertyChanged", self, "position", self:GetSavedPosition())
	end
end

function Bar:GetSavedPosition()
	local point = self.sets.point or 'CENTER'
	local x = tonumber(self.sets.x) or 0
	local y = tonumber(self.sets.y) or 0

	return point, x, y
end

-- gets the position of the bar relative to its parent
function Bar:GetRelativePosition()
	local scale = self.frame:GetScale() or 1
	local left = self.frame:GetLeft() or 0
	local top = self.frame:GetTop() or 0
	local right = self.frame:GetRight() or 0
	local bottom = self.frame:GetBottom() or 0

	local parent = self.frame:GetParent() or _G.UIParent
	local pwidth = parent:GetWidth() / scale
	local pheight = parent:GetHeight() / scale

	local x , y, point
	if left < (pwidth - right) and left < abs((left+right)/2 - pwidth/2) then
		x = left
		point = 'LEFT'
	elseif (pwidth - right) < abs((left + right)/2 - pwidth/2) then
		x = right - pwidth
		point = 'RIGHT'
	else
		x = (left+right)/2 - pwidth/2
		point = '';
	end

	if bottom < (pheight - top) and bottom < abs((bottom + top)/2 - pheight/2) then
		y = bottom
		point = 'BOTTOM' .. point
	elseif (pheight - top) < abs((bottom + top)/2 - pheight/2) then
		y = top - pheight
		point = 'TOP' .. point
	else
		y = (bottom + top)/2 - pheight/2
	end

	if point == '' then
		point = 'CENTER'
	end

	return point, x, y
end

function Bar:SaveRelativePosition()
	self:SetSavedPosition(self:GetRelativePosition())
end


-- Anchoring
Bar.stickyTolerance = 16

-- returns true if the this bar is attached to this bar
function Bar:UpdatePosition()
	local scale = self:GetScale()
	local pPoint, pRelativeTo, pRelativePoint, pX, pY = self.frame:GetPoint()

	self.frame:SetScale(scale)
	self.frame:SetPoint(self:GetSavedPosition())

	local nPoint, nRelativeTo, nRelativePoint, nX, nY = self.frame:GetPoint()

	local changed = not (
		pPoint == nPoint
		and pRelativeTo == nRelativeTo
		and pRelativePoint == nRelativePoint
		and floatsEqual(pX or 0, nX or 0)
		and floatsEqual(pY or 0, nY or 0)
	)

	if changed then
		Addon.Callbacks:Fire("OnBarPositionUpdated", self, nPoint, nRelativeTo, nRelativePoint, nX, nY)
	end
end

function Bar:SnapToGrid()
	local stuck = false
	local point, x, y = self:GetRelativePosition()
	local rTolerance = self.stickyTolerance / self:GetScale()

	if abs(x) <= rTolerance then
		stuck = true
		x = 0
	end

	if abs(y) <= rTolerance then
		stuck = true
		y = 0
	end

	if stuck then
		self.frame:ClearAllPoints()
		self.frame:SetPoint(point, x, y)
		return true
	end

	return false
end

-- Metafunctions
function Bar:Get(id)
	if id then
		return active[tonumber(id) or id]
	end
end

function Bar:GetAll()
	return pairs(active)
end

function Bar:ForAll(method, ...)
	for _, f in self:GetAll() do
		if type(method) == "string" then
			local action = f[method]
			if action then
				action(f, ...)
			end
		elseif type(method) == "function" then
			method(f, ...)
		end
	end
end

--takes a frameId, and performs the specified action on that frame
--this adds two special IDs, 'all' for all frames and '<number>-<number>' for a range of IDs
function Bar:ForIDs(id, method, ...)
	if id == 'all' then
		self:ForAll(method, ...)
	else
		local startID, endID = tostring(id):match('(%d+)-(%d+)')
		startID = tonumber(startID)
		endID = tonumber(endID)

		if startID and endID then
			if startID > endID then
				local t = startID
				startID = endID
				endID = t
			end

			for i = startID, endID do
				local f = self:Get(i)
				if f then
					local action = f[method]
					if action then
						action(f, ...)
					end
				end
			end
		else
			local f = self:Get(id)
			if f then
				local action = f[method]
				if action then
					action(f, ...)
				end
			end
		end
	end
end

-- exports
Addon.Bar = Bar