--[[
	fadeManager.lua
		Handles fading out frames when not moused over
		Necessary since using the blizzard fading functions can cause issues in combat
--]]

local Addon = _G[...]
local MouseOverWatcher = {}
local Timer_After = _G['C_Timer'].After
local watched = {}

local function isChildFocus(...)
	local focus = GetMouseFocus()
	for i = 1, select('#', ...) do
		if focus == select(i, ...) then
			return true
		end
	end

	for i = 1, select('#', ...) do
		local f = select(i, ...)
		if f:IsShown() and isChildFocus(f:GetChildren()) then
			return true
		end
	end

	return false
end

local function isDescendant(frame, ...)
	for i = 1, select('#', ...) do
		local f = select(i, ...)
		if frame == f then
			return true
		end
	end

	for i = 1, select('#', ...) do
		local f = select(i, ...)
		if isDescendant(frame, f:GetChildren()) then
			return true
		end
	end

	return false
end

--returns all frames docked to the given frame
local function isMouseOver(bar)
	local frame = bar.frame

	if frame:IsMouseOver(1, -1, -1, 1) then
		return (GetMouseFocus() == WorldFrame) or isChildFocus(frame:GetChildren())
	end

	if SpellFlyout and SpellFlyout:IsMouseOver(1, -1, -1, 1) and isDescendant(SpellFlyout:GetParent(), frame) then
		return true
	end
end

function MouseOverWatcher:OnBarInitialized(msg, bar)
	if bar:IsShown() and bar:GetFadeMultiplier() < 1 then
		self:Add(bar)
	else
		self:Remove(bar)
	end
end

function MouseOverWatcher:OnBarPropertyChanged(msg, bar, property)
	if not(property == "alpha" or property == "fadeMultiplier" or property == "show") then
		return
	end

	if bar:IsShown() and bar:GetFadeMultiplier() < 1 then
		self:Add(bar)
	else
		self:Remove(bar)
	end
end

function MouseOverWatcher:Update()
	for bar in pairs(watched) do
		if isMouseOver(bar) then
			if not bar.focused then
				bar.focused = true
				bar:UpdateAlpha()
			end
		else
			if bar.focused then
				bar.focused = nil
				bar:UpdateAlpha()
			end
		end
	end

	if next(watched) then
		self:RequestUpdate()
	end
end

function MouseOverWatcher:RequestUpdate()
	if not self.__Update then
		self.__Update = function()
			self.__Waiting = false
			self:Update()
		end
	end

	if not self.__Waiting then
		self.__Waiting = true
		Timer_After(0.15, self.__Update)
	end
end

function MouseOverWatcher:Add(bar)
	if not watched[bar] then
		watched[bar] = true

		bar.focused = isMouseOver(bar)
		bar:UpdateAlpha()

		self:RequestUpdate()
	end
end

function MouseOverWatcher:Remove(bar)
	if watched[bar] then
		watched[bar] = nil

		bar.focused = nil
		bar:UpdateAlpha()
	end
end

Addon.RegisterCallback(MouseOverWatcher, "OnBarInitialized")
Addon.RegisterCallback(MouseOverWatcher, "OnBarPropertyChanged")

--[[ exports ]]--

Addon.MouseOverWatcher = MouseOverWatcher