local AddonName, Addon = ...
local ParentAddonName = GetAddOnDependencies(AddonName)
local ParentAddon = _G[ParentAddonName]
local L = LibStub("AceLocale-3.0"):GetLocale(ParentAddonName)

local HSLColor = LibStub("LibHSLColor-1.0")

local KEYBOARD_MOVEMENT_INCREMENT = 1
local DRAG_FRAME_COLOR = HSLColor.CreateColor(120, 1, 0.3)

local DRAG_FRAME_BACKDROP = {
    bgFile = [[Interface\ChatFrame\ChatFrameBackground]],
    edgeFile = [[Interface\ChatFrame\ChatFrameBackground]],
    edgeSize = 2,
    tileSize = 12,
    insets = {left = 2, right = 2, top = 2, bottom = 2}
}

local round = _G.Round

-- frame that goes on each bar for movement/configuration
local DragFrame = Addon:CreateClass("Button")
local nextName = Addon:CreateNameGenerator("Overlay")
local overlays = {}

-- constructor
local function overlay_Create(parent)
    local frame = CreateFrame("Button", nextName(), parent)

    frame:EnableMouseWheel(true)
    frame:SetClampedToScreen(true)
    frame:SetBackdrop(DRAG_FRAME_BACKDROP)

    frame:SetNormalFontObject("GameFontGreen")
    frame:RegisterForClicks("AnyUp")
    frame:RegisterForDrag("LeftButton")
    frame:EnableKeyboard(true)

    return frame
end

function DragFrame:GetOrCreate(parent, owner)
    local f = self:Get(owner)

    if not f then
        f = self:New(parent, owner)
    end

    return f
end

function DragFrame:New(parent, owner)
    local frame = self:Bind(overlay_Create(parent))

    frame:SetScript("OnMouseDown", self.StartMoving)
    frame:SetScript("OnMouseUp", self.StopMoving)
    frame:SetScript("OnMouseWheel", self.OnMouseWheel)
    frame:SetScript("OnClick", self.OnClick)
    frame:SetScript("OnEnter", self.OnEnter)
    frame:SetScript("OnLeave", self.OnLeave)
    frame:SetScript("OnKeyUp", self.OnKeyUp)
    frame:SetScript("OnKeyDown", self.OnKeyDown)
    frame:SetScript("OnShow", self.OnShow)

    frame.owner = owner
    frame:SetText(owner.name or owner.id)
    frame:SetAllPoints(owner.frame)
    frame:SetFrameStrata("DIALOG")
    frame:UpdateColor()

    overlays[owner] = frame
    return frame
end

function DragFrame:OnClick(button)
    if button == "RightButton" then
        if IsModifierKeyDown() then
            self.owner:ToggleShown()
        else
            ParentAddon:ToggleBarOptionsMenu(self.owner)
        end
    elseif button == "MiddleButton" then
        self.owner:ToggleShown()
    end

    self:UpdateTooltip()
end

function DragFrame:OnEnter()
    if not self.isMouseOver then
        self.isMouseOver = true

        self:UpdateColor()
        self:UpdateTooltip()
        self:EnableArrowKeyMovement()
    end
end

function DragFrame:OnKeyDown(key)
    local handled = false

    if self.watchingKeyboardMovement then
        if key == "UP" then
            self:NudgeFrame(0, KEYBOARD_MOVEMENT_INCREMENT)
            handled = true
        elseif key == "DOWN" then
            self:NudgeFrame(0, -KEYBOARD_MOVEMENT_INCREMENT)
            handled = true
        elseif key == "LEFT" then
            self:NudgeFrame(-KEYBOARD_MOVEMENT_INCREMENT, 0)
            handled = true
        elseif key == "RIGHT" then
            self:NudgeFrame(KEYBOARD_MOVEMENT_INCREMENT, 0)
            handled = true
        end
    end

    self:SetPropagateKeyboardInput(not handled)
end

function DragFrame:OnLeave()
    if self.isMouseOver then
        self.isMouseOver = nil

        GameTooltip:Hide()
        self:UpdateColor()
        self:DisableArrowKeyMovement()
    end
end

function DragFrame:OnMouseWheel(delta)
    if IsModifierKeyDown() then
        self:CycleFocus(delta)
    else
        self:IncrementOpacity(delta)
    end
end

function DragFrame:OnShow()
    self:UpdateColor()
end

function DragFrame:StartMoving(button)
    if button == "LeftButton" then
        self.isMoving = true
        self.owner:StartMoving()

        if GameTooltip:IsOwned(self) then
            GameTooltip:Hide()
        end
    end
end

function DragFrame:StopMoving()
    if self.isMoving then
        self.isMoving = nil
        self.owner:StopMovingOrSizing()

        self:UpdateColor()
        self:UpdateTooltip()
    end
end

-- actions
function DragFrame:EnableArrowKeyMovement()
    self.watchingKeyboardMovement = true
    self:EnableKeyboard(true)
end

function DragFrame:DisableArrowKeyMovement()
    self.watchingKeyboardMovement = nil
    self:EnableKeyboard(false)
end

function DragFrame:NudgeFrame(dx, dy)
    local point, x, y = self.owner:GetRelativePosition()

    self.owner:ClearAnchor()
    self.owner:SetSavedPosition(point, round(x + dx), round(y + dy))
    self.owner:UpdatePosition()
end

function DragFrame:UpdateTooltip()
    local tooltip = _G.GameTooltip

    tooltip:SetOwner(self, "ANCHOR_BOTTOMLEFT")

    GameTooltip_SetTitle(tooltip, self.owner.name or self.owner.id)

    local description = self.owner.description
    if description then
        GameTooltip_AddNormalLine(tooltip, description .. "\n")
    end

    GameTooltip_AddInstructionLine(tooltip, L.ShowConfig)

    if self.owner:IsShown() then
        GameTooltip_AddInstructionLine(tooltip, L.HideBar)
    else
        GameTooltip_AddInstructionLine(tooltip, L.ShowBar)
    end

    local alphaHelp = L.SetAlpha:format(ceil(self.owner:GetAlpha() * 100))
    GameTooltip_AddInstructionLine(tooltip, alphaHelp)

    tooltip:Show()
end

do
    local framesUnderMouse = {}

    function DragFrame:CycleFocus(delta)
        wipe(framesUnderMouse)

        local currentIndex = 1
        local count = 0

        -- grab all frames under the mouse
        for _, frame in pairs(overlays) do
            if frame:IsMouseOver() then
                count = count + 1

                if GetMouseFocus() == frame then
                    currentIndex = count
                end

                tinsert(framesUnderMouse, frame)
            end
        end

        -- need at least two frames to cycle through
        if #framesUnderMouse < 2 then
            return
        end

        local nextIndex = currentIndex + delta

        if nextIndex > #framesUnderMouse then
            nextIndex = nextIndex % #framesUnderMouse
        elseif nextIndex <= 0 then
            nextIndex = #framesUnderMouse - nextIndex
        end

        framesUnderMouse[currentIndex]:SetFrameLevel(1)
        framesUnderMouse[nextIndex]:SetFrameLevel(100)
    end
end

function DragFrame:IncrementOpacity(delta)
    local oldAlpha = self.owner.sets and self.owner.sets.alpha or 1
    local newAlpha = min(max(oldAlpha + (delta * 0.05), 0), 1)

    -- round to fix floating point fun
    oldAlpha = math.floor(oldAlpha * 100 + 0.5) / 100
    newAlpha = math.floor(newAlpha * 100 + 0.5) / 100

    if newAlpha ~= oldAlpha then
        self.owner:SetAlpha(newAlpha)
        self:UpdateTooltip()
    end
end

-- coloring
function DragFrame:UpdateColor()
    self:SetBackdropColor(self:GetBackdropColor())
    self:SetBackdropBorderColor(self:GetBorderColor())
end

function DragFrame:GetColor()
    return DRAG_FRAME_COLOR.h, DRAG_FRAME_COLOR.s, DRAG_FRAME_COLOR.l
end

function DragFrame:GetBorderColor()
    local h, s, l = self:GetColor()
    local a = 1

    if self.isMouseOver then
        l = 0.7
    end

    if not self.owner:IsShown() then
        s = 0
    end

    return HSLColor.HSLAToRGBA(h, s, l, a)
end

function DragFrame:GetBackdropColor()
    return 0, 0, 0, 0.4
end

function DragFrame:Get(owner)
    return overlays[owner]
end

Addon.DragFrame = DragFrame