local AddonName, Addon = ...
local ParentAddonName = GetAddOnDependencies(AddonName)
local ParentAddon = _G[ParentAddonName]

function Addon:CreateClass(frameType, prototype)
    local class = CreateFrame(frameType); class:Hide()
    local class_mt = { __index = class }

    class.Bind = function(_, obj)
        return setmetatable(obj, class_mt)
    end

    if prototype then
        class.proto = prototype

        return setmetatable(class, {__index = prototype})
    end

    return class
end

-- returns a function that generates unique names for frames
-- in the format <AddonName>_<Prefix>[1, 2, ...]
function Addon:CreateNameGenerator(prefix)
    local id = 0

    return function()
        id = id + 1
        return ("%s_%s_%d"):format(AddonName, prefix, id)
    end
end

-- a thing to manage rendering stuff on the next frame
do
    local subscribers = {}
    local renderer = CreateFrame("Frame")
    renderer:Hide()

    renderer:SetScript(
        "OnUpdate",
        function(self)
            while next(subscribers) do
                table.remove(subscribers):OnRender()
            end

            self:Hide()
        end
    )

    function Addon:Render(frame)
        for _, f in pairs(subscribers) do
            if f == frame then
                return false
            end
        end

        table.insert(subscribers, 1, frame)
        renderer:Show()
        return true
    end
end

-- prevent updating until after a certain amount of time has passed
do
    local unpack = unpack
    local GetTime = _G.GetTime
    local now = function()
        return GetTime() * 1000
    end
    local after = _G.C_Timer.After

    function Addon:Debounce(func, delay)
        local args = {}
        local lastCall = 0

        local callback = function()
            if (now() - lastCall) >= delay then
                if #args > 0 then
                    func(unpack(args))
                else
                    func()
                end
            end
        end

        return function(...)
            wipe(args)

            for i = 1, select("#", ...) do
                args[i] = (select(i, ...))
            end

            lastCall = now()
            after(delay / 1000, callback)
        end
    end
end

-- exports
ParentAddon.Config = Addon
