local AddonName, Addon = ...
local ParentAddonName = GetAddOnDependencies(AddonName)
local ParentAddon = _G[ParentAddonName]
local L = LibStub("AceLocale-3.0"):GetLocale(ParentAddonName)

local SPACING = 12

local GamepadActionButton  = Addon:CreateClass('Frame')
do
    local function button_Pickup(self)
        ParentAddon.SecureEnv:PickupAction(self.kind, self.action)
        self:SetChecked(false)
    end

    local function button_OnEnter(self)
        GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
        GameTooltip_SetTitle(GameTooltip, self.name, HIGHLIGHT_FONT_COLOR)
    end

    local function button_OnLeave(self)
        if GameTooltip:IsOwned(self) then
            GameTooltip:Hide()
        end
    end

	function GamepadActionButton:New(options)
		local button = CreateFrame('CheckButton', nil, options.parent, 'ActionButtonTemplate')

		button.kind = options.kind
		button.action = options.action
        button.icon:SetTexture(options.icon)
        button.name = options.name

		if (options.count or 0) ~= 0 then
			button.Count:SetText(options.count)
		end

		button:RegisterForDrag("LeftButton", "RightButton")
		button:RegisterForClicks("AnyUp")
		button.Border:SetVertexColor(0, 1, 0, 0.35)
		button.NormalTexture:SetAlpha(0.5)

		button:SetScript("OnDragStart", button_Pickup)
        button:SetScript("OnClick", button_Pickup)
        button:SetScript("OnEnter", button_OnEnter)
        button:SetScript("OnLeave", button_OnLeave)

		return button
	end
end

local GamepadActionsPanel = Addon.OptionsUI:NewPanel(L.GamepadActionsPanelTitle)

-- force page display
local forcePageDropdown
do
	local forcedPage = -1

	local function setForcedPage(page)
		if page ~= -1 then
			forcedPage = page
			ParentAddon:ForcePage(page)
		end
	end

	local function getDropdownItems()
		local prevNumPages = nil
		local items = {}

		return function()
			local numPages = ParentAddon:NUM_GAMEPAD_PAGES()

			if prevNumPages ~= numPages then
				prevNumPages = numPages

				wipe(items)

				items = {
					{ value = -1, text = _G.DEFAULT },
				}

				for i = 1, numPages do
					items[i + 1] = {
						value = i,
						text = L.PageX:format(i)
					}
				end

				tinsert(items, {
					value = "override",
					text = L.Override
				})
			end

			return items
		end
	end

	local dropdown = GamepadActionsPanel:Add("Dropdown", {
		name = L.ShowPage,

		items = getDropdownItems(),

		get = function()
			return forcedPage
		end,

		set = function(_, value)
			setForcedPage(value)
		end
	})

	dropdown:HookScript("OnHide", function()
		setForcedPage(false)
	end)

	forcePageDropdown = dropdown
end

forcePageDropdown:SetPoint("TOPLEFT", 0, -SPACING)

local MovementActionsSection
do
    local f = CreateFrame("Frame", nil, GamepadActionsPanel)
    f:SetPoint('TOPLEFT', 0, -2)
    f:SetPoint('TOPRIGHT', 0, -2)

	local text = f:CreateFontString(nil, 'ARTWORK', 'GameFontNormal')
	text:SetJustifyH('LEFT')
	text:SetPoint('TOPLEFT')
	text:SetText(L.MovementActions)

	local border = f:CreateTexture(nil, 'ARTWORK')
	border:SetPoint('TOPLEFT', text, 'BOTTOMLEFT', 0, -2)
	border:SetPoint('RIGHT')
	border:SetHeight(1)
    border:SetColorTexture(0.3, 0.3, 0.3, 0.8)

    local prev
	for _, id in ParentAddon.GetMovementSlotIDs() do
        local name, icon = ParentAddon.GetMovementSlotInfo(id)

		local button = GamepadActionButton:New{
			parent = f,
			kind = "movement",
			name = name,
			action = id,
			icon = icon
        }

        if prev ~= nil then
            button:SetPoint("LEFT", prev, "RIGHT", SPACING, 0)
        else
            button:SetPoint("TOPLEFT", text, "BOTTOMLEFT", 0, -SPACING)
        end

        prev = button
    end

    local height = SPACING + 3
    height = height + text:GetHeight()
    height = height + prev:GetHeight()

    f:SetHeight(height)
    MovementActionsSection = f
end

local VehicleActionsSection
do
	local buttons = {}

	local f = CreateFrame("Frame", nil, GamepadActionsPanel)

	local function updateButtonDisplay()
		for _, button in pairs(buttons) do
			button.icon:SetDesaturated(not ParentAddon:UsingAsOverrideBar())
		end
	end

	f:SetScript("OnShow", updateButtonDisplay)

	local vehicleBarToggle = GamepadActionsPanel:Add('CheckButton', {
		name = L.ShowOverrideActionsOnGamePad,

		get = function()
			return ParentAddon:UsingAsOverrideBar()
		end,

		set = function(_, enable)
			ParentAddon:SetUseAsOverrideBar(enable)
			updateButtonDisplay()
		end
	})

	vehicleBarToggle:SetPoint('TOPLEFT', f, 'TOPLEFT', -2, 1)

	local border = f:CreateTexture(nil, 'ARTWORK')
	border:SetPoint('TOPLEFT', vehicleBarToggle, 'BOTTOMLEFT', 2, 2)
	border:SetPoint('RIGHT')
	border:SetHeight(1)
	border:SetColorTexture(0.3, 0.3, 0.3, 0.8)

	local actions = {
		{ BINDING_NAME_ACTIONBUTTON1, "overrideaction", 1, [[Interface\ICONS\INV_Misc_EngGizmos_27]], 1},
		{ BINDING_NAME_ACTIONBUTTON2, "overrideaction", 2, [[Interface\ICONS\INV_Misc_EngGizmos_27]], 2},
		{ BINDING_NAME_ACTIONBUTTON3, "overrideaction", 3, [[Interface\ICONS\INV_Misc_EngGizmos_27]], 3},
		{ BINDING_NAME_ACTIONBUTTON4, "overrideaction", 4, [[Interface\ICONS\INV_Misc_EngGizmos_27]], 4},
		{ BINDING_NAME_ACTIONBUTTON5, "overrideaction", 5, [[Interface\ICONS\INV_Misc_EngGizmos_27]], 5},
		{ BINDING_NAME_ACTIONBUTTON6, "overrideaction", 6, [[Interface\ICONS\INV_Misc_EngGizmos_27]], 6},
		{ BINDING_NAME_VEHICLEEXIT, "exitvehicle", "exitvehicle", ParentAddon.ExitVehicleSlot.GetIcon() },
	}

    local prev
	for _, data in ipairs(actions) do
		local name, kind, action, icon, count = unpack(data)

		local button = GamepadActionButton:New{
			parent = f,
			kind = kind,
			name = name,
			action = action,
			icon = icon,
			count = count
		}

        if prev ~= nil then
            button:SetPoint("LEFT", prev, "RIGHT", SPACING, 0)
        else
            button:SetPoint("TOPLEFT", vehicleBarToggle, "BOTTOMLEFT", 2, -(SPACING - 4))
        end

        prev = button
		tinsert(buttons, button)
    end

    local height = SPACING
	height = height + vehicleBarToggle:GetHeight()
    height = height + prev:GetHeight()

    f:SetHeight(height)
    VehicleActionsSection = f
end

local ActionPageActionSection
do
    local f = CreateFrame("Frame", nil, GamepadActionsPanel)

	local text = f:CreateFontString(nil, 'ARTWORK', 'GameFontNormal')
	text:SetJustifyH('LEFT')
	text:SetPoint('TOPLEFT', 0, 1)
	text:SetText("Action Pages")

	local border = f:CreateTexture(nil, 'ARTWORK')
	border:SetPoint('TOPLEFT', text, 'BOTTOMLEFT')
	border:SetPoint('RIGHT')
	border:SetHeight(1)
    border:SetColorTexture(0.3, 0.3, 0.3, 0.8)

    local prev
	for i = 1, 6 do
		local name = _G[("BINDING_NAME_ACTIONPAGE%d"):format(i)]

		local button = GamepadActionButton:New{
			parent = f,
			kind = "actionpage",
			name = name,
			action = i,
			icon = ParentAddon.ActionPageSlot.GetIcon(),
			count = i
		}

        if prev ~= nil then
            button:SetPoint("LEFT", prev, "RIGHT", SPACING, 0)
        else
            button:SetPoint("TOPLEFT", text, "BOTTOMLEFT", 0, -SPACING)
        end

        prev = button
    end

    local height = SPACING + 3
    height = height + text:GetHeight()
    height = height + prev:GetHeight()

    f:SetHeight(height)
    ActionPageActionSection = f
end

do
	local panels = {ActionPageActionSection, MovementActionsSection, VehicleActionsSection}

	for i, panel in ipairs(panels) do
		if i == 1 then
			panel:SetPoint("TOPLEFT", forcePageDropdown, "BOTTOMLEFT", 0, -SPACING)
			panel:SetPoint("RIGHT")
		else
			panel:SetPoint("TOPLEFT", panels[i - 1], "BOTTOMLEFT", 0, -SPACING)
			panel:SetPoint("TOPRIGHT", panels[i - 1], "BOTTOMRIGHT", 0, -SPACING)
		end
	end
end